/*	$Id: create.c,v 1.23 2005/08/22 21:48:34 phil Exp $ */

/* BLURB gpl

                           Coda File System
                              Release 6

          Copyright (c) 1987-2005 Carnegie Mellon University
                  Additional copyrights listed below

This  code  is  distributed "AS IS" without warranty of any kind under
the terms of the GNU General Public Licence Version 2, as shown in the
file  LICENSE.  The  technical and financial  contributors to Coda are
listed in the file CREDITS.

*/

///////////////////////////////////////////////////////////////////////////////
//
//  create.c -- Coda work items, opening cache files
//   
//  (C) Copyright 1998 OSR Open Systems Resources, Inc.
//  All Rights Reserved
//
//  OSR Open Systems Resources, Inc.
//  105 Route 101A Suite 19
//  Amherst, NH 03031  (603) 595-6500 FAX: (603) 595-6503
//
///////////////////////////////////////////////////////////////////////////////

#define THIS_FILE_TRACE __TRACE_CREATE__

//
// codawrap
// OSR call interface

#include <coda_types.h>
#include <coda.h>
#include <coda_nt.h>
#include <cnode.h>

typedef struct {
    //
    // Magic number
    //
    ULONG MagicNumber;
    
    //
    // This parameter is the standard "work item" information
    // used by the generic work queue package.
    //
    PIO_WORKITEM WorkItem;
    
    
    //
    // The next set of arguments are really just the parameters
    // needed to IoCreateFile.
    //
    UNICODE_STRING FileName;
    ULONG ObjectAttributes;
    HANDLE RelatedFileHandle; // could be null, must be valid in system context
    ULONG FileAttributes;
    ULONG Disposition;
    ULONG CreateOptions;
    
    //
    // Store the client security context.
    //
    PSECURITY_CLIENT_CONTEXT ClientSecurityContext;
    
    
    //
    // Finally, we have the "optional" return information.  This
    // includes an event (so the caller can wait for the operation
    // to complete.)  These parameters are all considered optional
    // and are only used if present.
    //
    KEVENT *event;
    HANDLE *FileHandle;
    PFILE_OBJECT *FileObject;
    IO_STATUS_BLOCK *IoSb;
    
} NULL_CREATE_WORK_ITEM, *PNULL_CREATE_WORK_ITEM;

//
// random number used for "magic #"
//

#define NULL_CREATE_WORK_ITEM_MAGIC 0x6f2c0abe
__inline BOOLEAN
NullFsdIsValidCreateWorkItem(PNULL_CREATE_WORK_ITEM CreateWorkItem) 
{
    return (CreateWorkItem->MagicNumber == NULL_CREATE_WORK_ITEM_MAGIC);
}


//
// CreateWorkRoutine
//
//   This routine is used to open a file in a worker thread and
//   hence always in some specific thread context.
//
// Inputs:
//   Parameter
//
// Outputs:
//   None.
//
// Returns:
//   VOID function
//
// Notes:
//   Using this routine provides us with a few advantages:
//    - It ensures we are running in a known (system) context for
//      manipulating file handles and file objects.
//    - It ensures we are running with known credentials (system)
//      for opening files.  Note: this works well for local
//      file systems, but not for remote file systems (where the
//      remote system won't trust "local system" credentials when
//      they arrive over the network. )  For remote systems one
//      should use impersonation tokens instead.
//
//  THE CALLER IS RESPONSIBLE FOR CLEANING UP THE WORK ITEM.
//
VOID CreateWorkRoutine(IN PDEVICE_OBJECT pdev, IN PVOID Parameter)
{
    PNULL_CREATE_WORK_ITEM createWorkItem = (PNULL_CREATE_WORK_ITEM) Parameter;
    OBJECT_ATTRIBUTES fileAttributes;
    NTSTATUS status;
    HANDLE newFileHandle;
    ULONG options;
    IO_STATUS_BLOCK iosb;
    PFILE_OBJECT newFileObject;
    OBJECT_HANDLE_INFORMATION newFileObjectHandleInfo;
    
    //
    // Make sure we got the right parameter
    //
    ASSERT(NullFsdIsValidCreateWorkItem(createWorkItem));
    
    //
    // We need to set up the object attributes
    //
    InitializeObjectAttributes(&fileAttributes,                   // attributes being initialized
                               &createWorkItem->FileName,         // name of file being opened
                               createWorkItem->ObjectAttributes,  // case insensitive, etc.
                               createWorkItem->RelatedFileHandle, // related handle, could be NULL
                               NULL);                             // security descriptor, not used
    
    //
    // Open the file.  We use IoCreateFile because it avoids the considerable overhead
    // of processing a software interrupt (via the Zw interface.)  This interface is also
    // more flexible than the various Zw* routines that sit on top of it.
    //
    
    status = IoCreateFile(&newFileHandle,    // new handle
                          MAXIMUM_ALLOWED,   // whatever we can get
                          &fileAttributes,
                          &iosb,
                          NULL,  // no initial allocation size
                          createWorkItem->FileAttributes,
                          FILE_SHARE_VALID_FLAGS, // share everything
                          createWorkItem->Disposition,
                          createWorkItem->CreateOptions,
                          NULL, // no ea buffer
                          0, // no ea buffer length
                          CreateFileTypeNone, // create file type
                          NULL, // no extra parameters
                          0); // no extra options
    
    if (!NT_SUCCESS(status)) {
	
        //
        // The operation failed.
        //
	
        if (createWorkItem->IoSb) {
	    
            createWorkItem->IoSb->Status = status;
	    
        }
	
        if (createWorkItem->event) {
	    
            KeSetEvent(createWorkItem->event, EVENT_INCREMENT, FALSE);
	    
        }
	
        //
        // Done.
        //
	
        return;
	
    }
    
    //
    // At this point the operation succeeded.  We need to convert the handle to
    // an object (since we really want objects.)
    //
    
    status = ObReferenceObjectByHandle(newFileHandle,
                                       MAXIMUM_ALLOWED,
                                       *IoFileObjectType,
                                       KernelMode,
                                       &newFileObject,
                                       &newFileObjectHandleInfo);
    
    if (!NT_SUCCESS(status)) {
	
        ZwClose(newFileHandle);
	
        if (createWorkItem->IoSb) {
	    
            createWorkItem->IoSb->Status = status;
	    
        }
	
        if (createWorkItem->event) {
	    
            KeSetEvent(createWorkItem->event, EVENT_INCREMENT, FALSE);
	    
        }
	
        //
        // Done!
        //
	
        return;
	
    }
    
    //
    // We now have the object, which is really what we wanted in the first place.
    // Finish up the operation and prepare to return to the caller.
    //
    if (createWorkItem->FileHandle) {
	
        *createWorkItem->FileHandle = newFileHandle;
	
    } else {
	
        //
        // Caller didn't want the file handle.  Close it.
        //
        
        (void) ZwClose(newFileHandle);
    }
    
    //
    // This is the file object we're returning.
    // NOTE THAT IT IS REFERENCE COUNTED!
    //
    if (createWorkItem->FileObject) {
	
        *createWorkItem->FileObject = newFileObject;
        
    } else {
	
        //
        // caller didn't want the file object. Decrement the reference
        // count.
        //
        ObDereferenceObject(newFileObject);
	
    }
    
    //
    //  This allows us to tell the caller what actually happened.
    //
    if (createWorkItem->IoSb) {
	
        *createWorkItem->IoSb = iosb;
	
    }
    
    //
    // If there was a close event, we note the close is done.  Since the
    // caller is responsible for cleanup, we can no longer use the work item
    // once we've indicated the close is done!
    //
    if (createWorkItem->event) {
	
        KeSetEvent(createWorkItem->event, EVENT_INCREMENT, FALSE);
        
    }
    
    //
    // Done!
    //
    
    return;
    
}


//
// CaptureCurrentSecurity
//
//  This routine is used to capture the security context of the current thread for
//  use via impersonation.
//
// Inputs:
//   None.
//
// Outputs:
//   None.
//
// Returns:
//   Pointer to current client security context
//
// Notes:
//   None.
//
PSECURITY_CLIENT_CONTEXT CaptureCurrentSecurity()
{
    SECURITY_QUALITY_OF_SERVICE securityQOS;
    PSECURITY_CLIENT_CONTEXT clientContext;
    NTSTATUS status;
    
    clientContext = coda_malloc(sizeof(SECURITY_CLIENT_CONTEXT));
    
    
    //
    // Check for allocation failure.
    //
    if (!clientContext) {
        return NULL;
    }
    
    RtlZeroMemory(clientContext, sizeof(SECURITY_CLIENT_CONTEXT));
    //
    // Initialize security QOS information
    //
    securityQOS.Length = sizeof(SECURITY_QUALITY_OF_SERVICE);
    securityQOS.ImpersonationLevel = SecurityImpersonation;
    securityQOS.ContextTrackingMode = SECURITY_DYNAMIC_TRACKING;
    securityQOS.EffectiveOnly = TRUE;
    
    //
    // Grab the client security information
    //
    status = SeCreateClientSecurity(PsGetCurrentThread(),
                                    &securityQOS,
                                    FALSE, // we use this info locally
                                    clientContext);
    
    if (!NT_SUCCESS(status)) {
	coda_free(clientContext);
        return NULL;
	
    }
    //
    // We have the security information.  Return it to the caller.
    //
    return clientContext;
}


//
// coda_open_container
//   This routine is used to open a file; it does this by building a work item
//   and waiting for the results.
//
// Inputs:
//   FileName - this is the name of the file to open
//   ObjectAttributes - anything "special" about this object
//   RelatedFileHandle - if this is being opened relative to some directory
//   FileAttributes - special file attributes
//   Disposition - creating a new file?  opening an existing file?
//   CreateOptions - special create/open options
//
// Outputs:
//   FileHandle - the newly opened file handle
//   FileObject - the newly opened file object
//
// Returns:
//  VOID function.
//
// Notes:
//  None.
//
VOID NullOpenFile(PUNICODE_STRING FileName,
                  ULONG ObjectAttributes,
                  HANDLE RelatedFileHandle,
                  ULONG FileAttributes,
                  ULONG Disposition,
                  ULONG CreateOptions,
                  OUT HANDLE *FileHandle,
                  OUT PFILE_OBJECT *FileObject,
                  OUT IO_STATUS_BLOCK *IoSb)
     
{
    NULL_CREATE_WORK_ITEM createWorkItem;
    LARGE_INTEGER createDelay;
    NTSTATUS status;
    
    //
    // Set up the work item
    //
    RtlZeroMemory(&createWorkItem, sizeof(createWorkItem));
    createWorkItem.MagicNumber = NULL_CREATE_WORK_ITEM_MAGIC;

    //
    // Allocate the event
    //
    createWorkItem.event = coda_malloc(sizeof(KEVENT));
    if (!createWorkItem.event) {
        //
        // Not enough memory?
        //
        IoSb->Status = STATUS_INSUFFICIENT_RESOURCES;
        IoSb->Information = 0;
        return;
    }
    
    
    //
    // Initialize the event for use
    //
    KeInitializeEvent(createWorkItem.event, NotificationEvent, FALSE);

    createWorkItem.WorkItem = IoAllocateWorkItem(coda_device_object);
    
    createWorkItem.FileName.Buffer =
	ExAllocatePoolWithTag(PagedPool, FileName->MaximumLength, 'NicN');
    if (NULL == createWorkItem.FileName.Buffer) {
        //
        // Out of memory
        //
        IoSb->Status = STATUS_INSUFFICIENT_RESOURCES;
        IoSb->Information = 0;
        return;
    }
    
    createWorkItem.FileName.Length = 0;
    createWorkItem.FileName.MaximumLength = FileName->MaximumLength;
    RtlCopyUnicodeString(&createWorkItem.FileName, FileName);
    createWorkItem.ObjectAttributes = ObjectAttributes;
    createWorkItem.RelatedFileHandle = RelatedFileHandle;
    createWorkItem.FileAttributes = FileAttributes;
    createWorkItem.Disposition = Disposition;
    createWorkItem.CreateOptions = CreateOptions;
    
    //
    //  We should capture the current credentials of the caller, but don't
    //  for now.
    //
    createWorkItem.ClientSecurityContext = CaptureCurrentSecurity();
    
    //
    // Output parameters
    //
    createWorkItem.FileHandle = FileHandle;
    createWorkItem.FileObject = FileObject;
    createWorkItem.IoSb = IoSb;
    
    //
    // If we're already in a system thread, we'll call directly.  This
    // avoids potential work queue thread deadlock as well as eliminates
    // an unnecessary thread context switch.
    //
    
    if (IoIsSystemThread(PsGetCurrentThread())) {
        CreateWorkRoutine(coda_device_object, (PVOID) &createWorkItem);
    } else {
        IoQueueWorkItem(createWorkItem.WorkItem,  CreateWorkRoutine,
			DelayedWorkQueue, (PVOID) &createWorkItem);
    }

    //
    // Wait for the operation to complete
    //
    
    createDelay.QuadPart = -100000000; // 10 seconds
    do {
	
        status = KeWaitForSingleObject(createWorkItem.event,
                                       Executive,
                                       KernelMode,
                                       FALSE,
                                       &createDelay);
	
        if (status == STATUS_TIMEOUT) {
	    
            //
            // Something is wrong.  Bail out.
            //
            // Note: there is a memory leak here; the alternative
            //       is to free this which could lead to a blue
            //       screen instead.
            //
            DbgPrint("NullOpenFile: Timed out waiting for create\n");
            IoSb->Status = STATUS_TIMEOUT;
	    IoSb->Information = 0;
	    
            return;
	    
        }
	
    } while (status);
    
    //
    // Cleanup our event.
    //
    coda_free(createWorkItem.event);

    // XXXX free the file name????    
    
    //
    // Done.
    //
    return;
    
}

int coda_get_fileobject(coda_filehandle *fh, int flags)
{
    coda_inode *inode = fh->fh_inode;
    int rc;
    NTSTATUS status;
    UNICODE_STRING UnicodeName;
    ANSI_STRING AnsiName;
    IO_STATUS_BLOCK iosb;
    HANDLE ContainerHandle;
    PFILE_OBJECT ContainerFileObject;
    OBJECT_ATTRIBUTES ObjAttr;


    ENTRY (D_FILE, "coda_get_fileobject");

    // Is it open and the object there?
    if (fh->fh_isopen ) {
	if (!inode->c_object) {
	    DbgPrint ("coda_get_fileobject: fh opened, but no object!\n");
	    return 1;  // Object must exist!
	}
	return 0;
    }

    LockInode (inode);

    // Set open flags, start with at least what was requested
    // in the call to get the file object!
    fh->fh_openflags = flags;
    if (fh->fh_lastacc &
	(FILE_READ_DATA | FILE_LIST_DIRECTORY | FILE_READ_ATTRIBUTES
	 | GENERIC_READ | FILE_EXECUTE | GENERIC_EXECUTE))
      fh->fh_openflags |= C_O_READ;
    if  (fh->fh_lastacc &
	 (FILE_WRITE_DATA | FILE_APPEND_DATA | GENERIC_WRITE ))
      fh->fh_openflags |= C_O_WRITE;
							     
    // get the name of the container file: Venus will prepend 
    // with user supplied \DosDevices\C:\Cygwin style cygwin root.
    rc = venus_open_by_path(&inode->c_fid, fh->fh_openflags,
			    inode->c_containername, fh->fh_uid);
    if ( rc ) {
	UnLockInode (inode);
	CDEBUG((D_FILE, "coda_get_fileobject: venus open failed with %d\n",
		rc));
	return rc;
    }

    inode->c_open_cnt++;

    // Get the file object 
    if (inode->c_object == NULL) {
    
	// make Unicode grrr...
	RtlInitUnicodeString(&UnicodeName, NULL);
	RtlInitAnsiString(&AnsiName, inode->c_containername); 
	status = RtlAnsiStringToUnicodeString(&UnicodeName, &AnsiName, TRUE);
	if ( !NT_SUCCESS(status) ) {
	    DbgPrint("coda_get_file_opject: Cannot convert name to Unicode\n");
	    UnLockInode (inode);
	    return status;
	}

	CDEBUG((D_FILE, "coda_get_fileobject: ansi name %s, "
		"Unicode name %S\n", inode->c_containername,
		UnicodeName.Buffer));
    
	NullOpenFile(&UnicodeName, //PUNICODE_STRING FileName
                 OBJ_CASE_INSENSITIVE,
                 NULL, // no related file
                 FILE_ATTRIBUTE_NORMAL, // atributes
                 FILE_OPEN, // Disposition
                 FILE_SYNCHRONOUS_IO_NONALERT, // CreateOptions
		 &ContainerHandle,  
                 &ContainerFileObject,
		 &iosb);
    
	RtlFreeUnicodeString(&UnicodeName);
    
	if (!NT_SUCCESS(iosb.Status)) {
	    UnLockInode (inode);
	    return (iosb.Status);
	}              
    
	inode->c_handle = ContainerHandle;
	inode->c_object = ContainerFileObject;
    }

    // Update fh information
    fh->fh_isopen = 1;

    UnLockInode (inode);
    CDEBUG((D_FILE, "coda_get_fileobject: Exiting, inode open count=%d\n",
	    inode->c_open_cnt));
    return 0;
}
